#include <iostream>
#include <string>
#include <vector>
#include <sstream>	// for std::ostringstream

#define NUM_OF_TEXTURES	5

typedef unsigned int   dword;
typedef unsigned short  word;
typedef unsigned char   byte;

struct Point2 {float x, y;};
struct Point3 {float x, y, z;};


byte GetByte(FILE* file)
{
	byte b = 0;
	fread(&b, 1, 1, file);
	return b;
}
word GetWord(FILE* file)
{
	word w = 0;
	fread(&w, 2, 1, file);
	return w;
}
dword GetDword(FILE* file)
{
	dword dw = 0;
	fread(&dw, 4, 1, file);
	return dw;
}
void Skip(FILE* file, int l)
{
	fseek(file, l, SEEK_CUR);
}
int ReadData(FILE* file, void* dataOut, int dataLen)
{
	return static_cast<int>(fread(dataOut, 1, dataLen, file));
}
std::string ReadStr(FILE* file)
{
	static char temp[1024] = {0};

	word length = GetWord(file);
	fread(temp, 1, length, file);
	std::string result;
	result.assign(temp, length);
	return result;
}



int main(int argc, char** argv)
{
	if (argc < 3)
	{
		std::cout << "Usage: ds2md_to_obj.exe input_model.ds2md output_model.obj" << std::endl;
		return -1;
	}

	std::string inputModel = argv[1];
	std::string outputModel = argv[2];

	FILE* fileSrc = fopen(inputModel.c_str(), "rb");
	if (!fileSrc)
	{
		std::cout << "Could not open " << inputModel << std::endl;
		return -1;
	}

	FILE* fileDst = fopen(outputModel.c_str(), "wb");
	if (!fileSrc)
	{
		fclose(fileSrc);
		std::cout << "Could not create " << outputModel << std::endl;
		return -1;
	}

	// file ID
	std::string fileID = ReadStr(fileSrc);
	// model version
	std::string version = ReadStr(fileSrc);
	// model name
	std::string modelName = ReadStr(fileSrc);
	// unknown byte
	byte oneByte = GetByte(fileSrc);
	// material name
	std::string materialName = ReadStr(fileSrc);

	// try to find texture
	std::string textureName;
	for (int i = 0; i < NUM_OF_TEXTURES; ++i)
	{
		std::string texture = ReadStr(fileSrc);
		if (textureName.empty() && texture != "notexture")
			textureName = texture;
	}

	// skip unknown 33 bytes
	Skip(fileSrc, 33);

	// read indexes
	dword numIndexes = GetDword(fileSrc);
	std::vector<word> indexes(numIndexes);
	ReadData(fileSrc, &indexes.front(), numIndexes * sizeof(word));

	// read vertexes data
	word numVerts = GetWord(fileSrc);
	std::vector<Point3> positions(numVerts);
	std::vector<Point3> normals(numVerts);
	std::vector<Point2> texcoords(numVerts);

	ReadData(fileSrc, &positions.front(), numVerts * sizeof(Point3));
	ReadData(fileSrc, &normals.front(), numVerts * sizeof(Point3));
	ReadData(fileSrc, &texcoords.front(), numVerts * sizeof(Point2));

	// and now write obj file
	std::ostringstream	stringBuilder;
	stringBuilder << "# Generated from You Are Empty model file " << inputModel << std::endl;
	stringBuilder << "#     statistics: " << numIndexes << " indexes, " << numVerts << " vertexes" << std::endl << std::endl;

	// write model name
	stringBuilder << "g " << modelName << std::endl << std::endl;

	// write positions
	for (std::vector<Point3>::const_iterator it = positions.begin(); it != positions.end(); ++it)
		stringBuilder << "v " << it->x << " " << it->y << " " << it->z << std::endl;

	stringBuilder << std::endl;

	// write texcoords
	for (std::vector<Point2>::const_iterator it = texcoords.begin(); it != texcoords.end(); ++it)
		stringBuilder << "vt " << it->x << " " << 1.0f - it->y << std::endl;	// for 3DSMax we flip y coordinate of texcoords

	stringBuilder << std::endl;

	// write normals
	for (std::vector<Point3>::const_iterator it = normals.begin(); it != normals.end(); ++it)
		stringBuilder << "vn " << it->x << " " << it->y << " " << it->z << std::endl;

	stringBuilder << std::endl;

	// write indexes
	for (dword i = 0; i < numIndexes; i += 3)
	{
		const word a = indexes[i + 0] + 1;
		const word b = indexes[i + 1] + 1;
		const word c = indexes[i + 2] + 1;

		stringBuilder << "f " << a << "/" << a << "/" << a <<
						  " " << b << "/" << b << "/" << b <<
						  " " << c << "/" << c << "/" << c <<  std::endl;
	}

	fwrite(stringBuilder.str().c_str(), stringBuilder.str().length(), 1, fileDst);
	fflush(fileDst);

	fclose(fileSrc);
	fclose(fileDst);
}
